// fudgeish.hpp
#ifndef FUDGEISH_HPP_INCLUDED
#define FUDGEISH_HPP_INCLUDED

#include <list>
#include <vector>
#include <iostream>
#include <string>

/**
License: Public Domain
Author: stephan@s11n.net

namespace fudge contains objects representing some of the core
aspects of the Fudge roleplaying system.

Most (99.99%) of the naming and logic here were inspired by Fudge
(http://www.fudgerpg.com). Indeed, the code is for processing common
Fudge-style operations.

*/

namespace fudge {

        /**
           Rating lists the Fudge-standard levels used for rating
           almost everything: strength, difficulty ratings,
           etc.
        */
        enum Rating
                {
                RatingTerrible = -3,
                RatingPoor = -2,
                RatingMediocre = -1,
                RatingFair = 0,
                RatingGood = 1,
                RatingGreat = 2,
                RatingSuperb = 3
                };

        /**
           MassScale: 

           Combat damage is mediated largely by combatants' relative
           mass scales...

           This list is largely taken from:
           http://plm.snow.org/fudge/scale.html

           It is by no means authoritative - interpret as you see fit
           - but most of it is backed by solid math (see the above
           URL) and works well with the default Fudge combat
           mechanics.
        */
        enum MassScale
                {
                Scale_Electron = -181,
                Scale_Proton = -162,
                Scale_Cat = -7,
                Scale_Wolf = -1,
                Scale_Human = 0,
                Scale_BlackBear = 3,
                Scale_Tiger = 4,
                Scale_SmallCar = 5,
                Scale_Horse = 6,
                Scale_SmallHelicopter = 8,
                Scale_SmallTruck = 10,
                Scale_Elephant = 11,
                Scale_FighterJet = 12,
                Scale_Bus = 13,
                Scale_Lorey = 14,
                Scale_LightTank = 15,
                Scale_MBT = 17,
                Scale_NavalDestroyer_Small = 24,
                Scale_NavalDestroyer = 28,
                Scale_WW2BattleShip_Small = 29,
                Scale_NuclearSubmarine = 30,
                Scale_WW2BattleShip = 33,
                Scale_AircraftCarrier = 34,
                Scale_LargeAsteroid = 108,
                Scale_Moon = 120,
                Scale_Terra = 131,
                Scale_Sol = 162,
                Scale_SolarSystem = 180, // COMPLETE GUESS!!!
                Scale_Universe = 290
                };

        /**
           Rolls a contest of skill against the given difficulty rating
           and returns the relative result.

           Example:

           Good Strength needs a Great result to break down a
           heavy wooden door. Calling:

           rollContest( Great, Good );

           would return (Good - Great) + 4dF, which would be -1
           (Mediocre) most of the time.

         */
        int rollContest( int difficultyRating, int skillRating );

        /**
           Returns a string for a given Fudge rating. Thus passing it
           0 returns "Fair". Passing it numbers larger than Superb
           return "Superb+N", and smaller than Terrible returns
           "Terrible-N".
         */
        std::string ratingName( int rating );

	/**
	   Returns a number in the set (-1,0,1). It automatically
	   seeds the RNG the first time it is called.
	*/
	short rollDF();

        /**
           dF represents a single "Fudge Die" (aka, dF).
        */
        class dF
        {
        public:
                /**
                   Creates a new dF and calls side( whichside ).
                 */
                dF( short whichside = 0 );
                virtual ~dF(){}

                /**
                   Returns the currently-"shown" side (one of: -1, 0, 1).
                 */
                inline short side() const { return this->m_side; }

                /**
                   Sets the current side to one of -1, 0, +1.
                   If s < 0 then it is set to -1. If s > 1 then it is set to 1.
                 */
                inline void side( short s ) { this->m_side = ( (s == 0) ? 0 : ( s < 0 ? -1 : 1 ) ); }

                /**
                   Returns one of:

                   	+ - 0

                   representing the current side. The 0 (zero)
                   /really/ should be a blank space, because that's
                   what the physical dF have, but that's impossible to
                   differentiate from other spaces in s7ed
                   data. :/
                 */
                char symbol() const;

                /**
                   Adds one of "-", "0", or "+" to os.
                 */
                friend std::ostream & operator<<( std::ostream & os, const dF & hp );

                /**
                   Reads one character from is and translates -, 0 or
                   + as -1, 0 or 1, respectively. Any other character
                   is treated as 0. hp's side() is set accordingly.
                 */
                friend std::istream & operator>>( std::istream & is, dF & hp );
                       
                /**
                   Sets the current side to a randomly selected one in
                   the set (-1,0,+1) and returns the current side.
                */
                short roll();

        private:
                short m_side;
                // i'd include color here, too, since i have so many different colors of 
                // Fudge dice, but it's representation is too implementation-specific :/
        };

        /**
        DiceBag represents a bag of dice from which you can roll
        multiple dice at one time. Okay, it should perhaps be a cup of
        dice, but it's not because my Fudge dice came with a cool
        little carrying bag, and thus i associate them with a bag. :)
        (Thanks, Ann!)

        A DiceBag is also intended to be used as "dice manager". i.e.,
        you give each player his own dice bag and that is the bag from
        which he rolls. A dice bag is, of course, re-rollable, so you
        can re-use the same static dice bag for all rolls of a given
        number of dice.
        */
        class DiceBag
        {
        public:
                typedef std::vector<dF> list_type;
                typedef list_type::iterator iterator;
                typedef list_type::const_iterator const_iterator;

                /**
                   Creates a bag filled with dicecount dice.
                 */
                DiceBag( int dicecount = 4 );
                virtual ~DiceBag(){}

                /**
                   Rolls howmany dice and returns the total via the return result.
                   The exact dice rolled are returned via the passed-in thedice.

                   todo: templatize it to take arbitrary container types.
                */
                static int roll( int howmany, list_type & thedice );


                /**
                   Similar to the static roll() except that this one gives you
                   no access to the actual dice results. Use dice() to get at
                   those. It returns the total result of the dice roll.
                */
                int roll();

                /**
                   Replaces the current dice with count dice.
                 */
                void diceCount( int count );
                /**
                   Returns the number of dice in the bag.
                */
                int diceCount() const;


// might be interesting: (but comparing the int results is SO much simpler!)
//                 bool operator==( const DiceBag &rhs );
//                 bool operator<( const DiceBag &rhs );
//                 bool operator>( const DiceBag &rhs );

                /**
                   Returns the value set by the last call to
                   roll().
                 */
                inline int previousRoll() const { return this->m_lastroll; }

                /**
                   The list of dice representing the previous
                   roll(). This list is cleared on any call to roll(),
                   so copy the data if you need it and don't trust any
                   existing iterators after a call to roll().
                 */
                list_type & dice();
                const list_type & dice() const;

                /**
                   Iterate over your dice with begin() and end(), STL-style.
                 */
                const_iterator begin() const { return this->m_dice.begin(); }
                iterator begin() { return this->m_dice.begin(); }
                const_iterator end() const { return this->m_dice.end(); }
                iterator end() { return this->m_dice.end(); }

                /**
                   Inserts the results of bag's last roll, in a
                   human-readable format, into os.
                */
                friend std::ostream & operator<<( std::ostream & os, const DiceBag & bag );

                /**
                   Primarily intended as a helper for operator&lt;&lt;( ostream &, DiceBag & ).
                */
                friend std::ostream & operator<<( std::ostream & os, const DiceBag::list_type & dice );


        private:
                list_type m_dice;
                int m_lastroll;

        };


        /**
           WoundChart represents a Fudge-standard-style hit chart,with
           parameter-definable sizes for setting the number of wounds
           boxes for any given wound severity.
        */
        class WoundChart
        {
        public:
                typedef std::vector<int> vector_type;
                typedef vector_type::iterator iterator;

                /**
                   Severities represents a range of wound severities.
                   Note that Dead is a special case, and not part of
                   the core rules.  It's here to make it simpler for
                   computers to determine if a Fudge combatant is dead
                   or not. Death is a rather shady thing in Fudge, and
                   computers don't handle this lack of rules
                   gracefully.

                   Developer note:

                   It is absolutely imperitive that these enum values
                   not changes unless you also change every vector[x]
                   and map[x] call in this code. They are used as
                   vector subscripts and map keys all over the
                   place. Normally i hate this type of inflexibility
                   but i figure the Fudge core rules have stood solid
                   for somewhere around ten years now and the system
                   has worked well, so... i've avoid the temptation to
                   code something more generic :/.
                 */
                enum Severities
                        { 
                        NoEffect = 0,
                        Scratch = 1,
                        Hurt = 2,
                        VeryHurt = 3,
                        Incap = 4,
                        NearDeath = 5,
                        Dead = 6
                        };

                /**
                   Creates a new wounds track. The passed-in
                   parameters correspond to Scratch .. NearDead. Their
                   values are the number of wound boxes this chart has
                   for that hit severity. The default values are, of
                   course, from the default Fudge wounds chart.
                 */
                WoundChart( int MaxScratch = 3, int MaxHurt=2, int MaxVeryHurt = 1, int MaxIncap = 1, int MaxNearDeath = 1 );
                virtual ~WoundChart(){}

                /**
                   Returns true if this chart has been Scratched.
                 */
                inline bool isScratched() { return this->m_hits[Scratch] > 0; }

                /**
                   Returns true if this chart has a Hurt mark.
                 */
                inline bool isHurt() const { return this->m_hits[Hurt] > 0; }

                /**
                   Returns true if this chart has any Very Hurt wounds.
                 */
                inline bool isVeryHurt() const { return this->m_hits[VeryHurt] > 0; }

                /**
                   Returns true if this chart has any Incapacitating wounds.
                 */
                inline bool isIncap() const { return this->m_hits[Incap] > 0; }

                /**
                   i'll leave you guessing what this does...
                */
                inline bool isNearDeath() const { return this->m_hits[NearDeath] > 0; }

                /**
                   Clears all damage, resetting the chart to its
                   default state. ("It was just a scratch!")
                 */
                void clear();

                /**
                   Returns true if all of this object's wound boxes are "checked off".
                 */
                bool isFull();

                /**
                   Returns true if any damage has spilled over "past NearDeath", regardless
                   of whether the wound track is full.
                 */
                bool isDead();

                // todo? API for manipulating number of wound boxes.

                /**
                   Fills one wound box of Severity severity.  If all
                   boxes of that type are filled then the wound will
                   bump up to the next available wound level (if
                   possible).

                   See migrateWounds() for important information on
                   how wounds are "bumped up" when slots are full at a
                   given severity level.
                 */
                int wound( int severity );

                
                /**
                   If migrateWounds() is set to true then all wound()s
                   will get bumped up to the next severity level once
                   a given severity level is full. i.e., once you run
                   out of Scratch slots, the next Scratch becomes a
                   Hurt.

                   Defaults to true.
                 */
                inline void migrateWounds( bool b ) { this->m_migratewounds = b; }

                /**
                   Returns true if this object "migrates wounds", as
                   described in migrateWounds(bool), else returns
                   false.
                 */
                inline bool migrateWounds() const { return this->m_migratewounds; }

                /**
                   Returns the name of the given severity. e.g.,
                   passing it WoundChart::Scratch returns "Scratch".
                   For technical reasons names are returned without
                   spaces, e.g. "No effect" is "NoEffect" and "Very
                   Hurt" is "VeryHurt". Sorry about that. :/
                 */
                static std::string woundName( int severity );

                /**
                   Sends a simple rendering of this object's state to os.
                */
                friend std::ostream & operator<<( std::ostream & os, const WoundChart &rhs );

                /**
                   The list of maximum wound boxes for each
                   Severity. This exists mainly for serialization
                   purposes.
                */
                vector_type & boxes() { return this->m_boxes; }
                const vector_type & boxes() const { return this->m_boxes; }

		/**
		   The list of current hits to each hit wound chart
		   location. This exists mainly for serialization
		   purposes.
		*/
                vector_type & hits() { return this->m_hits; }
                const vector_type & hits() const { return this->m_hits; }

		/**
		   A list enumerating how many hits of damage
		   translate to which wound type. This exists mainly
		   for serialization purposes.
		*/
                vector_type & damageChart() { return this->m_dmgchart; }
                const vector_type & damageChart() const { return this->m_dmgchart; }

                /**
                   Applies dmg to this object's wounds box.
                   It uses severity() to determine what type of wound
                   dmg represents.

                   It returns the wound level applied by the damage
                   (NoEffect .. Dead). Pass its return result to
                   woundName() to get a string.
                */
                
                virtual int applyDamage( int dmg );

                /**
                   Returns the fudge::Rating severity for dmg damage points.

                   Uses (rather unobjectively) the Fudge-default rules
                   for translating damage points to hits.

                   Uses the damage table set via damageConversion().

                   It does not account for any Mass Scale effects, nor
                   anything situation-specific: those are assumed to
                   be applied to dmg before passing it here.
                */
                virtual int severity( int dmg );

                /**
                   Sets the damage --> severity conversion
                   chart. Defaults to Fudge standard usage, with the
                   addition that MinDead is also set (because a
                   computer can't make arbitrary decicions about
                   whether "near death" is dead enough).
                   
                   If MinDead == -1 then MinDead is set to MinND plus
                   the difference of MinND and MinIncap. That is:
                   MinND + MinND - MinIncap

                   The intention is to follow the pattern set by the caller.

                   If MinDead is set to an explicit number then that number is used.

                   Results are undefined if each parameter's value is
                   not greater than the one before it, except the
                   special case of MinDead == -1.

                   Default conversion is Fudge standard:

                   1-2 = Scratch

                   3-4 = Hurt

                   5-6 = Very Hurt

                   7-8 = Incapacitated

                   9+ = Near Death

                   That's all standard Fudge. Here's where it differs:

                   11+ = flags Death, so isDead() will finally return true.

                   Again, this is to accomodate computers.
                */
                void damageConversion( int MinScratch, int MinHurt, int MinVH, int MinIncap, int MinND, int MinDead = -1 );

        private:
                vector_type m_hits;
                vector_type m_boxes;
                vector_type m_dmgchart;
                bool m_migratewounds;
                void init();

        };
   

        /**
           A Buff represents an in-use item or feature/enhancement
           which offers ODF or DDF enhancements during combat. The
           exact interpretation of a buff is client-dependent.
        */
        class Buff
        {
        public:
                Buff(const std::string &n = "Unnamed Buff") : m_odf(0),m_ddf(0),m_scale(Scale_Human),m_range(0),m_count(1),m_ablative(false),m_flags(0),m_name(n)
                {
                        this->init();
                }

                Buff( const std::string &n, int o, int d, int s=fudge::Scale_Human ) : m_odf(o),m_ddf(d),m_scale(s),m_range(0),m_count(1),m_ablative(false),m_flags(0),m_name(n)
                {
                        this->init();
                }

                virtual ~Buff(){}

                /**
                   If a Buff may only be used intermittently, then this should be overriden
                   to return false in some cases.
                 */
                virtual bool mayUse() { return true; }

                /**
                   odf() Offensive Damage Factors. i.e., "damage".
                   The core Fudge rules call it this, thus the odd name.
                 */
                void odf( int x ) { this->m_odf = x; }
                int odf() const { return this->m_odf; }

                /**
                   ddf() Defensive Damage Factors. i.e., "armor", "damage reduction", or whatever you like to call it.
                   The core Fudge rules call it this.
                 */
                void ddf( int x ) { this->m_ddf = x; }
                int ddf() const { return this->m_ddf; }

                /**
                   Fudge scale, though which scale it is is up to the
                   client. It will normally apply to the mass
                   scale. Note that it is okay for a MOB and his buffs
                   to have different scales: imagine a huge cybertank
                   with huge [high-scale] armor-smashing weapons
                   alongside [low-scale] peasant-shredder weapons.
                 */
                void scale( int x ) { this->m_scale = x; }
                int scale() const { return this->m_scale; }

                /**
                   Every Buff's got a name...
                 */
                void name( const std::string & x ) { this->m_name = x; }
                const std::string & name() const { return this->m_name; }

                /**
                   range(): not always applicable, and always of
                   arguable interpretation.  The returned value may or
                   may not be in the range (-3)..(+3). Its
                   interpretation is client-specific.
                 */
                void range( int x ) { this->m_range = x; }
                int range() const { return this->m_range; }


                /**
                   This can be used to represent "charged" buffs,
                   which go away after some number of uses. See
                   ablative().
                 */
                void count( int x ) { this->m_count = x; }
                int count() const { return this->m_count; }

                /**
                   A flag which tells clients that it should reduce this buff's count()
                   when it uses this buff. What constitues as "use", of course,
                   is client-specific. Ablative is false by default.
                 */
                void ablative( bool x ) { this->m_ablative = x; }
                bool ablative() const { return this->m_ablative; }

                friend std::ostream & operator<<( std::ostream & os, const Buff & ca );
                
                /**
                   Client-specific flags, like flagging offensive or defensive buffs.
                 */
                enum Flags { None = 0x00,
                             Ablative = 0x01, /* todo: port ablative to use flags. */
                             Offensive = 0x02,
                             Defensive = 0x04
                };

                inline void flags( int x ) { this->m_flags = x; }
                inline int flags() const { return this->m_flags; }
                
                /**
                   offensive/defensive() are convenience methods for setting the Offensive
                   and Defensive flags().

                   Rather arbitrary logic in the ctors sets offensive() to true if ODF != 0.
                   Similarly, defensive() gets set if DDF != 0.
                 */
                bool offensive() const { return this->flags() & Buff::Offensive; }
                void offensive( bool b )
                {
                        this->flags( this->flags() | (b ? Buff::Offensive : !Buff::Offensive) );
                }
                bool defensive() const { return (this->flags() & Buff::Defensive); }
                void defensive( bool b )
                {
                        this->flags( this->flags() | (b ? Buff::Defensive : !Buff::Defensive) );
                }


        private:
                int m_odf;
                int m_ddf;
                int m_scale; // note that this is the scale of the intended weilder.
                int m_range;
                int m_count;
                bool m_ablative;
                unsigned int m_flags;
                std::string m_name;

                void init()
                {
                        this->offensive( this->m_odf != 0 );
                        this->defensive( this->m_ddf != 0 );
                }
        };

} // namespace fudge


#endif // FUDGEISH_HPP_INCLUDED
